"""
Known stages to defer jobs to within the OpenAleph stack.

See [Settings][openaleph_procrastinate.settings.DeferSettings]
for configuring queue names and tasks.

Conventions / common pattern: Tasks are responsible to explicitly defer
following tasks. This defer call is not conditional but happens always, but
actually deferring happens in this module and is depending on runtime settings
(see below).

Example:
    ```python
    from openaleph_procrastinate import defer

    @task(app=app)
    def analyze(job: DatasetJob) -> None:
        result = analyze_entities(job.load_entities())
        # defer to index stage
        defer.index(app, job.dataset, result)
    ```

To disable deferring for a service, use environment variable:

For example, to disable indexing entities after ingestion, start the
`ingest-file` worker with this config: `OPENALEPH_INDEX_DEFER=0`
"""

from typing import Any, Iterable

from banal import ensure_dict
from followthemoney.proxy import EntityProxy

from openaleph_procrastinate.app import App
from openaleph_procrastinate.model import DatasetJob, Job
from openaleph_procrastinate.settings import DeferSettings

tasks = DeferSettings()


def get_priority(data: dict[str, Any], default: int) -> int:
    return data.get("priority") or default


def ingest(
    app: App, dataset: str, entities: Iterable[EntityProxy], **context: Any
) -> None:
    """
    Defer a new job for `ingest-file`.
    It will only deferred if `OPENALEPH_INGEST_DEFER=1` (the default)

    Args:
        app: The procrastinate app instance
        dataset: The ftm dataset or collection
        entities: The file or directory entities to ingest
        context: Additional job context
    """
    if tasks.ingest.defer:
        job = DatasetJob.from_entities(
            dataset=dataset,
            queue=tasks.ingest.queue,
            task=tasks.ingest.task,
            entities=entities,
            **context,
        )
        priority = get_priority(context, tasks.ingest.get_priority())
        job.defer(app, priority)


def analyze(
    app: App, dataset: str, entities: Iterable[EntityProxy], **context: Any
) -> None:
    """
    Defer a new job for `ftm-analyze`
    It will only deferred if `OPENALEPH_ANALYZE_DEFER=1` (the default)

    Args:
        app: The procrastinate app instance
        dataset: The ftm dataset or collection
        entities: The entities to analyze
        context: Additional job context
    """
    if tasks.analyze.defer:
        job = DatasetJob.from_entities(
            dataset=dataset,
            queue=tasks.analyze.queue,
            task=tasks.analyze.task,
            entities=entities,
            dehydrate=True,
            **context,
        )
        priority = get_priority(context, tasks.analyze.get_priority())
        job.defer(app, priority)


def index(
    app: App, dataset: str, entities: Iterable[EntityProxy], **context: Any
) -> None:
    """
    Defer a new job to index into OpenAleph
    It will only deferred if `OPENALEPH_INDEX_DEFER=1` (the default)

    Args:
        app: The procrastinate app instance
        dataset: The ftm dataset or collection
        entities: The entities to index
        context: Additional job context
    """
    if tasks.index.defer:
        job = DatasetJob.from_entities(
            dataset=dataset,
            queue=tasks.index.queue,
            task=tasks.index.task,
            entities=entities,
            dehydrate=True,
            **context,
        )
        priority = get_priority(context, tasks.index.get_priority())
        job.defer(app, priority)


def reindex(app: App, dataset: str, **context: Any) -> None:
    """
    Defer a new job to reindex into OpenAleph
    It will only deferred if `OPENALEPH_REINDEX_DEFER=1` (the default)

    Args:
        app: The procrastinate app instance
        dataset: The ftm dataset or collection
        context: Additional job context
    """
    if tasks.reindex.defer:
        job = DatasetJob(
            dataset=dataset,
            queue=tasks.reindex.queue,
            task=tasks.reindex.task,
            payload={"context": ensure_dict(context)},
        )
        priority = get_priority(context, tasks.reindex.get_priority())
        job.defer(app, priority)


def xref(app: App, dataset: str, **context: Any) -> None:
    """
    Defer a new job to xref into OpenAleph
    It will only deferred if `OPENALEPH_XREF_DEFER=1` (the default)

    Args:
        app: The procrastinate app instance
        dataset: The ftm dataset or collection
        context: Additional job context
    """
    if tasks.xref.defer:
        job = DatasetJob(
            dataset=dataset,
            queue=tasks.xref.queue,
            task=tasks.xref.task,
            payload={"context": ensure_dict(context)},
        )
        priority = get_priority(context, tasks.xref.get_priority())
        job.defer(app, priority)


def load_mapping(app: App, dataset: str, **context: Any) -> None:
    """
    Defer a new job to load_mapping into OpenAleph
    It will only deferred if `OPENALEPH_LOAD_MAPPING_DEFER=1` (the default)

    Args:
        app: The procrastinate app instance
        dataset: The ftm dataset or collection
        context: Additional job context
    """
    if tasks.load_mapping.defer:
        job = DatasetJob(
            dataset=dataset,
            queue=tasks.load_mapping.queue,
            task=tasks.load_mapping.task,
            payload={"context": ensure_dict(context)},
        )
        priority = get_priority(context, tasks.load_mapping.get_priority())
        job.defer(app, priority)


def flush_mapping(app: App, dataset: str, **context: Any) -> None:
    """
    Defer a new job to flush_mapping into OpenAleph
    It will only deferred if `OPENALEPH_FLUSH_MAPPING_DEFER=1` (the default)

    Args:
        app: The procrastinate app instance
        dataset: The ftm dataset or collection
        context: Additional job context
    """
    if tasks.flush_mapping.defer:
        job = DatasetJob(
            dataset=dataset,
            queue=tasks.flush_mapping.queue,
            task=tasks.flush_mapping.task,
            payload={"context": ensure_dict(context)},
        )
        priority = get_priority(context, tasks.flush_mapping.get_priority())
        job.defer(app, priority)


def export_search(app: App, **context: Any) -> None:
    """
    Defer a new job to export_search into OpenAleph
    It will only deferred if `OPENALEPH_EXPORT_SEARCH_DEFER=1` (the default)

    Args:
        app: The procrastinate app instance
        context: Additional job context
    """
    if tasks.export_search.defer:
        job = Job(
            queue=tasks.export_search.queue,
            task=tasks.export_search.task,
            payload={"context": ensure_dict(context)},
        )
        priority = get_priority(context, tasks.export_search.get_priority())
        job.defer(app, priority)


def export_xref(app: App, dataset: str, **context: Any) -> None:
    """
    Defer a new job to export_xref into OpenAleph
    It will only deferred if `OPENALEPH_EXPORT_XREF_DEFER=1` (the default)

    Args:
        app: The procrastinate app instance
        context: Additional job context
    """
    if tasks.export_xref.defer:
        job = DatasetJob(
            dataset=dataset,
            queue=tasks.export_xref.queue,
            task=tasks.export_xref.task,
            payload={"context": ensure_dict(context)},
        )
        priority = get_priority(context, tasks.export_xref.get_priority())
        job.defer(app, priority)


def update_entity(app: App, dataset: str, **context: Any) -> None:
    """
    Defer a new job to update_entity into OpenAleph
    It will only deferred if `OPENALEPH_UPDATE_ENTITY_DEFER=1` (the default)

    Args:
        app: The procrastinate app instance
        dataset: The ftm dataset or collection
        context: Additional job context
    """
    if tasks.update_entity.defer:
        job = DatasetJob(
            dataset=dataset,
            queue=tasks.update_entity.queue,
            task=tasks.update_entity.task,
            payload={"context": ensure_dict(context)},
        )
        priority = get_priority(context, tasks.update_entity.get_priority())
        job.defer(app, priority)


def prune_entity(app: App, dataset: str, **context: Any) -> None:
    """
    Defer a new job to prune_entity into OpenAleph
    It will only deferred if `OPENALEPH_PRUNE_ENTITY_DEFER=1` (the default)

    Args:
        app: The procrastinate app instance
        dataset: The ftm dataset or collection
        context: Additional job context
    """
    if tasks.prune_entity.defer:
        job = DatasetJob(
            dataset=dataset,
            queue=tasks.prune_entity.queue,
            task=tasks.prune_entity.task,
            payload={"context": ensure_dict(context)},
        )
        priority = get_priority(context, tasks.prune_entity.get_priority())
        job.defer(app, priority)


def cancel_dataset(app: App, dataset: str, **context: Any) -> None:
    """
    Defer a new job to cancel a dataset processing in OpenAleph
    It will only deferred if `OPENALEPH_CANCEL_DATASET_DEFER=1` (the default)

    Args:
        app: The procrastinate app instance
        dataset: The ftm dataset or collection
        context: Additional job context
    """
    if tasks.cancel_dataset.defer:
        job = DatasetJob(
            dataset=dataset,
            queue=tasks.cancel_dataset.queue,
            task=tasks.cancel_dataset.task,
            payload={"context": ensure_dict(context)},
        )
        priority = get_priority(context, tasks.cancel_dataset.get_priority())
        job.defer(app, priority)


def transcribe(
    app: App, dataset: str, entities: Iterable[EntityProxy], **context: Any
) -> None:
    """
    Defer a new job for `ftm-transcribe`
    It will only deferred if `OPENALEPH_TRANSCRIBE_DEFER=1` (the default)

    Args:
        app: The procrastinate app instance
        dataset: The ftm dataset or collection
        entities: The file entities to ingest
        context: Additional job context
    """
    if tasks.transcribe.defer:
        job = DatasetJob.from_entities(
            dataset=dataset,
            queue=tasks.transcribe.queue,
            task=tasks.transcribe.task,
            entities=entities,
            dehydrate=True,
            **context,
        )
        priority = get_priority(context, tasks.transcribe.get_priority())
        job.defer(app, priority)


def geocode(
    app: App, dataset: str, entities: Iterable[EntityProxy], **context: Any
) -> None:
    """
    Defer a new job for `ftm-geocode`
    It will only deferred if `OPENALEPH_GEOCODE_DEFER=1` (the default)

    Args:
        app: The procrastinate app instance
        dataset: The ftm dataset or collection
        entities: The entities to geocode
        context: Additional job context
    """
    if tasks.geocode.defer:
        job = DatasetJob.from_entities(
            dataset=dataset,
            queue=tasks.geocode.queue,
            task=tasks.geocode.task,
            entities=entities,
            **context,
        )
        priority = get_priority(context, tasks.geocode.get_priority())
        job.defer(app, priority)


def resolve_assets(
    app: App, dataset: str, entities: Iterable[EntityProxy], **context: Any
) -> None:
    """
    Defer a new job for `ftm-assets`
    It will only deferred if `OPENALEPH_ASSETS_DEFER=1` (the default)

    Args:
        app: The procrastinate app instance
        dataset: The ftm dataset or collection
        entities: The entities to resolve assets for
        context: Additional job context
    """
    if tasks.assets.defer:
        job = DatasetJob.from_entities(
            dataset=dataset,
            queue=tasks.assets.queue,
            task=tasks.assets.task,
            entities=entities,
            **context,
        )
        priority = get_priority(context, tasks.assets.get_priority())
        job.defer(app, priority)



"""sql queries for status aggregation and cancel jobs"""

# HELPER VARS #
JOBS = "procrastinate_jobs"
EVENTS = "procrastinate_events"
CTE_JOB_STATUS = "cte_job_status"  # CTE

SYSTEM_DATASET = "__system__"
DEFAULT_BATCH = "default"

COLUMNS = "dataset, batch, queue_name, task_name, status"
MAX_ID = "MAX(id) AS max_id"
MIN_ID = "MIN(id) AS min_id"

# FILTERS #
F_DATASET = "(%(dataset)s::varchar IS NULL OR dataset = %(dataset)s)"
F_BATCH = "(%(batch)s::varchar IS NULL OR batch = %(batch)s)"
F_QUEUE = "(%(queue)s::varchar IS NULL OR queue_name = %(queue)s)"
F_TASK = "(%(task)s::varchar IS NULL OR task_name = %(task)s)"
F_STATUS = "(%(status)s::procrastinate_job_status IS NULL OR status = %(status)s)"
F_ALL_ANDS = " AND ".join((F_DATASET, F_BATCH, F_QUEUE, F_TASK, F_STATUS))

# FOR INITIAL SETUP #
GENERATED_FIELDS = f"""
ALTER TABLE {JOBS}
ADD COLUMN IF NOT EXISTS dataset TEXT GENERATED ALWAYS AS (
    COALESCE(args->>'dataset', '{SYSTEM_DATASET}')
) STORED;

ALTER TABLE {JOBS}
ADD COLUMN IF NOT EXISTS batch TEXT GENERATED ALWAYS AS (
    COALESCE(args->>'batch', '{DEFAULT_BATCH}')
) STORED;
"""

# INDEXES FOR STATUS QUERY #
INDEXES = f"""
CREATE INDEX IF NOT EXISTS idx_{JOBS}_args
ON {JOBS} USING GIN (args);

CREATE INDEX IF NOT EXISTS idx_{JOBS}_dataset
ON {JOBS} (dataset);

CREATE INDEX IF NOT EXISTS idx_{JOBS}_batch
ON {JOBS} (batch);

CREATE INDEX IF NOT EXISTS idx_{JOBS}_task
ON {JOBS} (task_name);

CREATE INDEX IF NOT EXISTS idx_{JOBS}_status
ON {JOBS} (status);

CREATE INDEX IF NOT EXISTS idx_{JOBS}_grouping
ON {JOBS} (dataset, batch, queue_name, task_name, status);

CREATE INDEX IF NOT EXISTS idx_{EVENTS}_job_id_at
ON {EVENTS} (job_id, at);
"""


# QUERY JOB STATUS #
# query status aggregation, optional filtered for dataset. this is quite
# expensive, make sure the additional indexes and generated fields exist.
# this returns result rows with these values in its order:
# dataset,batch,queue_name,task_name,status,jobs count,first event,last event
STATUS_SUMMARY = f"""
WITH {CTE_JOB_STATUS} AS (
    SELECT {COLUMNS}, COUNT(*) AS jobs, {MIN_ID}, {MAX_ID} FROM {JOBS}
    WHERE {F_DATASET}
    GROUP BY {COLUMNS} ORDER BY {COLUMNS}
)
SELECT {COLUMNS},
MAX(jobs), MIN(e1.at), MAX(e2.at) FROM {CTE_JOB_STATUS}
LEFT JOIN {EVENTS} e1 ON min_id = e1.job_id
LEFT JOIN {EVENTS} e2 ON max_id = e2.job_id
GROUP BY {COLUMNS} ORDER BY {COLUMNS}
"""

# only return status aggregation for active datasets
STATUS_SUMMARY_ACTIVE = f"""
WITH {CTE_JOB_STATUS} AS (
    SELECT {COLUMNS}, COUNT(*) AS jobs, {MIN_ID}, {MAX_ID} FROM {JOBS} j1
    WHERE {F_DATASET}
    AND EXISTS (
        SELECT 1 FROM {JOBS} j2
        WHERE j2.dataset = j1.dataset
        AND j2.status IN ('todo', 'doing')
    )
    GROUP BY {COLUMNS} ORDER BY {COLUMNS}
)
SELECT {COLUMNS},
MAX(jobs), MIN(e1.at), MAX(e2.at) FROM {CTE_JOB_STATUS}
LEFT JOIN {EVENTS} e1 ON min_id = e1.job_id
LEFT JOIN {EVENTS} e2 ON max_id = e2.job_id
GROUP BY {COLUMNS} ORDER BY {COLUMNS}
"""


ALL_JOBS = f"""
SELECT id, status, args FROM {JOBS} WHERE id IN (
SELECT job_id FROM {EVENTS}
WHERE (at BETWEEN %(min_ts)s AND %(max_ts)s)
AND {F_ALL_ANDS}
)
"""

# CANCEL OPS #
# they follow the logic from here:
# https://github.com/procrastinate-org/procrastinate/blob/main/procrastinate/sql/schema.sql
# but alter the table in batch instead of running it one by one per job id.
# This is equivalent to the function `procrastinate_cancel_job_v1` with delete=true,abort=true
CANCEL_JOBS = f"""
DELETE FROM {JOBS} WHERE status = 'todo'
AND {F_DATASET} AND {F_BATCH} AND {F_QUEUE} AND {F_TASK};

UPDATE {JOBS} SET abort_requested = true, status = 'cancelled'
WHERE status = 'todo'
AND {F_DATASET} AND {F_BATCH} AND {F_QUEUE} AND {F_TASK};

UPDATE {JOBS} SET abort_requested = true
WHERE status = 'doing'
AND {F_DATASET} AND {F_BATCH} AND {F_QUEUE} AND {F_TASK};
"""



"""
This is temporary and should use the procrastinate Django models at one point in
the future
"""

from datetime import datetime
from typing import Any, Generator, LiteralString, TypeAlias

import psycopg
from anystore.functools import weakref_cache as cache
from anystore.logging import get_logger
from anystore.util import Took, mask_uri
from psycopg.errors import UndefinedTable

from openaleph_procrastinate.app import make_app
from openaleph_procrastinate.manage import sql
from openaleph_procrastinate.model import AnyJob, DatasetJob, EntityJob, Status
from openaleph_procrastinate.settings import OpenAlephSettings
from openaleph_procrastinate.tasks import unpack_job

RowType: TypeAlias = str | int | datetime | dict[str, Any]
Rows: TypeAlias = Generator[tuple[RowType, ...], None, None]
Jobs: TypeAlias = Generator[AnyJob | EntityJob, None, None]


class Db:
    """Get a db manager object for the current procrastinate database uri"""

    def __init__(self, uri: str | None = None) -> None:
        self.settings = OpenAlephSettings()
        if self.settings.in_memory_db:
            raise RuntimeError("Can't use in-memory database")
        self.uri = uri or self.settings.procrastinate_db_uri
        self.log = get_logger(__name__, uri=mask_uri(self.uri))

    def configure(self) -> None:
        """Create procrastinate tables and schema (if not exists) and add our
        index optimizations (if not exists)"""
        if self.settings.in_memory_db:
            return
        app = make_app(sync=True)
        with app.open():
            if not app.check_connection():
                self.log.info("Configuring procrastinate database schema ...")
                app.schema_manager.apply_schema()
        self.log.info("Configuring indexes and generated fields ...")
        with Took() as t:
            self._execute(sql.GENERATED_FIELDS)
            self._execute(sql.INDEXES)
            self.log.info("Configuring done.", took=t.took)

    def iterate_status(
        self,
        dataset: str | None = None,
        batch: str | None = None,
        queue: str | None = None,
        task: str | None = None,
        status: Status | None = None,
        active_only: bool | None = True,
    ) -> Rows:
        """
        Iterate through aggregated job status summary

        Each row is an aggregation over
        `dataset,batch,queue_name,task_name,status` and includes jobs count,
        timestamp first event, timestamp last event

        Args:
            dataset: The dataset to filter for
            batch: The job batch to filter for
            queue: The queue name to filter for
            task: The task name to filter for
            status: The status to filter for
            active_only: Only include "active" datasets (at least 1 job in
                'todo' or 'doing')

        Yields:
            Rows a tuple with the fields in this order:
                dataset, batch, queue_name, task_name, status, jobs count,
                timestamp first event, timestamp last event
        """
        if active_only:
            query = sql.STATUS_SUMMARY_ACTIVE
        else:
            query = sql.STATUS_SUMMARY
        yield from self._execute_iter(
            query,
            dataset=dataset,
            batch=batch,
            queue=queue,
            task=task,
            status=status,
        )

    def iterate_jobs(
        self,
        dataset: str | None = None,
        batch: str | None = None,
        queue: str | None = None,
        task: str | None = None,
        status: Status | None = None,
        min_ts: datetime | None = None,
        max_ts: datetime | None = None,
        flatten_entities: bool | None = False,
    ) -> Jobs:
        """
        Iterate job objects from the database by given criteria.

        Args:
            dataset: The dataset to filter for
            batch: The job batch to filter for
            queue: The queue name to filter for
            task: The task name to filter for
            status: The status to filter for
            min_ts: Start timestamp (earliest event found in `procrastinate_events`)
            max_ts: End timestamp (latest event found in `procrastinate_events`)
            flatten_entities: If true, yield a job for each entity found in the source job

        Yields:
            Iterator of [Job][openaleph_procrastinate.model.Job]
        """

        min_ts = min_ts or datetime(1970, 1, 1)
        max_ts = max_ts or datetime.now()
        params = {
            "dataset": dataset,
            "min_ts": min_ts.isoformat(),
            "max_ts": max_ts.isoformat(),
            "batch": batch,
            "queue": queue,
            "task": task,
            "status": status,
        }
        for id, status_, data in self._execute_iter(sql.ALL_JOBS, **params):
            data["id"] = id
            data["status"] = status_
            job = unpack_job(data)
            if flatten_entities and isinstance(job, DatasetJob):
                has_entities = False
                for entity in job.get_entities():
                    if entity.id:
                        has_entities = True
                        yield EntityJob(**data, entity_id=entity.id)
                if not has_entities:
                    yield job
            else:
                yield job

    def cancel_jobs(
        self,
        dataset: str | None = None,
        batch: str | None = None,
        queue: str | None = None,
        task: str | None = None,
    ) -> None:
        """
        Cancel jobs by given criteria.

        Args:
            dataset: The dataset to filter for
            batch: The job batch to filter for
            queue: The queue name to filter for
            task: The task name to filter for
        """

        self._execute(
            sql.CANCEL_JOBS, dataset=dataset, batch=batch, queue=queue, task=task
        )

    def _execute_iter(self, q: LiteralString, **params: str | None) -> Rows:
        with psycopg.connect(self.settings.procrastinate_db_uri) as connection:
            with connection.cursor() as cursor:
                cursor.execute(q, dict(params))
                while rows := cursor.fetchmany(100_000):
                    yield from rows

    def _execute(self, q: LiteralString, **params: str | None) -> None:
        with psycopg.connect(self.settings.procrastinate_db_uri) as connection:
            with connection.cursor() as cursor:
                for query in q.split(";"):
                    cursor.execute(query, dict(params))

    def _destroy(self) -> None:
        """Destroy all data (used in tests)"""
        self.log.warning("🔥 Deleting all procrastinate data 🔥")
        for table in (
            "procrastinate_jobs",
            "procrastinate_workers",
            "procrastinate_events",
        ):
            try:
                self._execute(f"TRUNCATE {table} RESTART IDENTITY CASCADE")
            except UndefinedTable:
                pass


@cache
def get_db(uri: str | None = None) -> Db:
    """Get a globally cached `db` instance"""
    settings = OpenAlephSettings()
    uri = uri or settings.procrastinate_db_uri
    return Db(uri)



import functools
import random
from typing import Any, Callable

from anystore.logging import get_logger
from procrastinate.app import App

from openaleph_procrastinate.exceptions import ErrorHandler
from openaleph_procrastinate.model import AnyJob, DatasetJob, Job

log = get_logger(__name__)


def unpack_job(data: dict[str, Any]) -> AnyJob:
    """Unpack a payload to a job"""
    with ErrorHandler(log):
        if "dataset" in data:
            return DatasetJob(**data)
        return Job(**data)


def task(app: App, **kwargs):
    # https://procrastinate.readthedocs.io/en/stable/howto/advanced/middleware.html
    def wrap(func: Callable[..., None]):
        def _inner(*job_args, **job_kwargs):
            # turn the json data into the job model instance
            job = unpack_job(job_kwargs)
            func(*job_args, job)

        # need to call to not register tasks twice (procrastinate complains)
        wrapped_func = functools.update_wrapper(_inner, func, updated=())
        # call the original procrastinate task decorator with additional
        # configuration passed through
        return app.task(**kwargs)(wrapped_func)

    return wrap


class _Priorities:
    """
    Use different priority buckets in tasks:

    Example:
        ```python
        from openaleph_procrastinate.tasks import Priorities

        defer_task(payload, priority=Priorities.MEDIUM)
        ```
    """

    MAX = 100

    @property
    def ANY(self) -> int:
        return random.randint(1, 100)

    @property
    def LOW(self) -> int:
        return random.randint(1, 50)

    @property
    def MEDIUM(self) -> int:
        return random.randint(50, 70)

    @property
    def HIGH(self) -> int:
        return random.randint(70, 90)

    @property
    def USER(self) -> int:
        return random.randint(90, 99)


Priorities = _Priorities()



import random
from datetime import UTC, datetime, timedelta
from pathlib import Path
from typing import Any, ContextManager, Generator, Iterable, Literal, Self, TypeAlias

from anystore.logging import BoundLogger, get_logger
from anystore.store.virtual import VirtualIO
from anystore.util import clean_dict
from banal import ensure_dict
from followthemoney import model
from followthemoney.proxy import EntityProxy
from ftmq.store.fragments.loader import BulkLoader
from pydantic import BaseModel, ConfigDict, computed_field

from openaleph_procrastinate import helpers
from openaleph_procrastinate.app import App, run_sync_worker
from openaleph_procrastinate.settings import (
    MAX_PRIORITY,
    MIN_PRIORITY,
    OpenAlephSettings,
)
from openaleph_procrastinate.util import make_checksum_entity

settings = OpenAlephSettings()


def get_priority() -> int:
    return random.randint(MIN_PRIORITY, MAX_PRIORITY)


class EntityFileReference(BaseModel):
    """
    A file reference (via `content_hash`) to a servicelayer file from an entity
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    dataset: str
    content_hash: str
    entity: EntityProxy

    def open(self: Self) -> ContextManager[VirtualIO]:
        """
        Open the file attached to this job
        """
        return helpers.open_file(self.dataset, self.content_hash)

    def get_localpath(self: Self) -> ContextManager[Path]:
        """
        Get a temporary path for the file attached to this job
        """
        return helpers.get_localpath(self.dataset, self.content_hash)


Status = Literal[
    "todo", "doing", "succeeded", "failed", "aborted", "aborting", "cancelled"
]


class JobModel(BaseModel):
    """
    A job with arbitrary payload
    """

    queue: str
    task: str
    payload: dict[str, Any] = {}

    # from procrastinate table:
    id: int | None = None
    status: Status | None = None
    scheduled_at: datetime | None = None


class Job(JobModel):
    """
    A job with arbitrary payload
    """

    @property
    def context(self) -> dict[str, Any]:
        """Get the context from the payload if any"""
        return ensure_dict(self.payload.get("context")) or {}

    @property
    def log(self) -> BoundLogger:
        return get_logger(name="openaleph.job", queue=self.queue, task=self.task)

    def defer(self: Self, app: App, priority: int | None = None) -> None:
        """Defer this job"""
        self.log.debug("Deferring ...", payload=self.payload)
        data = clean_dict(self.model_dump(mode="json"))
        app.configure_task(
            name=self.task, queue=self.queue, priority=priority or get_priority()
        ).defer(**data)
        if settings.debug:
            # option to change synchronousness during test runtime
            _settings = OpenAlephSettings()
            if _settings.procrastinate_sync:
                # run worker synchronously (for testing)
                run_sync_worker(app)


class DatasetJob(Job):
    """
    A job with arbitrary payload bound to a `dataset`.
    The payload always contains an iterable of serialized `EntityProxy` objects
    in the `entities` key. It may contain other payload context data in the
    `context` key.

    There are helpers for accessing archive files or loading entities.
    """

    dataset: str
    batch: str | None = None

    @property
    def log(self) -> BoundLogger:
        return get_logger(
            name=f"openaleph.job.{self.dataset}",
            dataset=self.dataset,
            queue=self.queue,
            task=self.task,
            batch=self.batch,
        )

    def get_writer(self: Self) -> ContextManager[BulkLoader]:
        """Get the writer for the dataset of the current job"""
        return helpers.entity_writer(self.dataset)

    def get_entities(self) -> Generator[EntityProxy, None, None]:
        """
        Get the entities from the payload
        """
        assert "entities" in self.payload, "No entities in payload"
        for data in self.payload["entities"]:
            yield model.get_proxy(data)

    def load_entities(self: Self) -> Generator[EntityProxy, None, None]:
        """Load the entities from the store to refresh it to the latest data"""
        assert "entities" in self.payload, "No entities in payload"
        for data in self.payload["entities"]:
            yield helpers.load_entity(self.dataset, data["id"])

    # Helpers for file jobs that access the servicelayer archive

    def get_file_references(self) -> Generator[EntityFileReference, None, None]:
        """
        Get file references per entity from this job

        Example:
            ```python
            # process temporary file paths
            for reference in job.get_file_references():
                with reference.get_local_path() as path:
                    subprocess.run(["command", "-i", str(path)])
                # temporary path will be cleaned up when leaving context

            # process temporary file handlers
            for reference in job.get_file_references():
                with reference.open() as handler:
                    do_something(handler.read())
                # temporary path will be cleaned up when leaving context
            ```

        Yields:
            The file references
        """
        for entity in self.get_entities():
            for content_hash in entity.get("contentHash", quiet=True):
                yield EntityFileReference(
                    dataset=self.dataset, entity=entity, content_hash=content_hash
                )

    @classmethod
    def from_entities(
        cls,
        dataset: str,
        queue: str,
        task: str,
        entities: Iterable[EntityProxy],
        dehydrate: bool | None = False,
        **context: Any,
    ) -> Self:
        """
        Make a job to process entities for a dataset

        Args:
            dataset: Name of the dataset
            queue: Name of the queue
            task: Python module path of the task
            entities: Entities
            dehydrate: Reduce entity payload to only a reference (tasks should
                re-fetch these entities from the store if they need more data)
            context: Job context
        """
        if dehydrate:
            entities_ = (make_checksum_entity(e, quiet=True) for e in entities)
            entities = (e for e in entities_ if e is not None)
        return cls(
            dataset=dataset,
            queue=queue,
            task=task,
            batch=context.pop("batch", None),
            payload={
                "entities": [e.to_dict() for e in entities],
                "context": ensure_dict(context),
            },
        )


class EntityJob(JobModel):
    dataset: str
    entity_id: str


AnyJob: TypeAlias = Job | DatasetJob


class StatusCounts(BaseModel):
    todo: int = 0
    doing: int = 0
    succeeded: int = 0
    failed: int = 0
    aborted: int = 0
    aborting: int = 0
    cancelled: int = 0

    min_ts: datetime | None = None
    max_ts: datetime | None = None

    @computed_field
    @property
    def remaining_time(self) -> timedelta | None:
        if self.finished and self.min_ts and self.max_ts:
            took = self.max_ts - self.min_ts
            remaining = (took.seconds / self.finished) * self.todo
            return timedelta(seconds=remaining)

    @computed_field
    @property
    def took(self) -> timedelta | None:
        max_ts = self.max_ts or datetime.now(UTC)
        if self.finished and self.min_ts:
            return max_ts - self.min_ts
        if self.min_ts:
            return datetime.now(UTC) - self.min_ts

    @computed_field
    @property
    def total(self) -> int:
        return sum(
            (
                self.todo,
                self.doing,
                self.succeeded,
                self.failed,
                self.aborted,
                self.aborting,
                self.cancelled,
            )
        )

    @computed_field
    @property
    def active(self) -> int:
        return self.todo + self.doing

    @computed_field
    @property
    def finished(self) -> int:
        return self.succeeded + self.failed + self.aborted + self.cancelled

    def is_active(self) -> bool:
        return self.active > 0

    def is_running(self) -> bool:
        return self.doing > 0

    def add_child_stats(self, child: "StatusCounts") -> None:
        self.todo += child.todo
        self.doing += child.doing
        self.succeeded += child.succeeded
        self.failed += child.failed
        self.aborted += child.aborted
        self.cancelled += child.cancelled
        if child.min_ts:
            if not self.min_ts or (self.min_ts and self.min_ts > child.min_ts):
                self.min_ts = child.min_ts
        if child.max_ts:
            if not self.max_ts or (self.max_ts and self.max_ts < child.max_ts):
                self.max_ts = child.max_ts


class TaskStatus(StatusCounts):
    name: str


class QueueStatus(StatusCounts):
    name: str
    tasks: list[TaskStatus] = []


class BatchStatus(StatusCounts):
    name: str
    queues: list[QueueStatus] = []


SYSTEM_DATASET = "__system__"


class DatasetStatus(StatusCounts):
    name: str
    batches: list[BatchStatus] = []

    def is_system(self) -> bool:
        return self.name == SYSTEM_DATASET



from collections import defaultdict
from typing import Generator

from openaleph_procrastinate.manage.db import get_db
from openaleph_procrastinate.model import (
    SYSTEM_DATASET,
    BatchStatus,
    DatasetStatus,
    QueueStatus,
    TaskStatus,
)

DEFAULT_BACH = "default"


def _gather_status(
    dataset: str | None = None, active_only: bool | None = True
) -> Generator[DatasetStatus, None, None]:
    db = get_db()
    tree = lambda: defaultdict(tree)  # noqa: E731
    data = tree()
    for (
        dataset_name,
        batch,
        queue,
        task,
        status,
        jobs,
        min_ts,
        max_ts,
    ) in db.iterate_status(dataset, active_only=active_only):
        data[dataset_name][batch][queue][task]["counts"][status] = jobs
        data[dataset_name][batch][queue][task]["min_ts"][status] = min_ts
        data[dataset_name][batch][queue][task]["max_ts"][status] = max_ts

    for dataset_name, batches in data.items():
        dataset_status = DatasetStatus(name=dataset_name or SYSTEM_DATASET)
        for batch, queues in batches.items():
            batch_status = BatchStatus(name=batch or DEFAULT_BACH)
            for queue, tasks in queues.items():
                queue_status = QueueStatus(name=queue)
                for task, stats in tasks.items():
                    min_ts = min(stats["min_ts"].values())
                    max_ts = min(stats["max_ts"].values())
                    task_status = TaskStatus(
                        name=task,
                        **stats["counts"],
                        min_ts=min_ts,
                        max_ts=max_ts,
                    )
                    queue_status.add_child_stats(task_status)
                    queue_status.tasks.append(task_status)
                batch_status.add_child_stats(queue_status)
                batch_status.queues.append(queue_status)
            dataset_status.add_child_stats(batch_status)
            dataset_status.batches.append(batch_status)
        yield dataset_status


def get_status(active_only: bool | None = True) -> Generator[DatasetStatus, None, None]:
    yield from _gather_status(active_only=active_only)


def get_dataset_status(dataset: str, active_only: bool | None = True) -> DatasetStatus:
    for status in _gather_status(dataset, active_only):
        return status
    return DatasetStatus(name=dataset)



from openaleph_procrastinate.manage.db import get_db


def cancel_jobs(
    dataset: str | None = None,
    batch: str | None = None,
    queue: str | None = None,
    task: str | None = None,
) -> None:
    db = get_db()
    db.cancel_jobs(dataset, batch, queue, task)



from anystore.logging import get_logger
from followthemoney.proxy import EntityProxy

log = get_logger(__name__)


def make_stub_entity(e: EntityProxy) -> EntityProxy | None:
    """
    Reduce an entity to its ID and schema
    """
    if not e.id:
        log.warning("Entity has no ID!")
        return
    return EntityProxy.from_dict(
        {"id": e.id, "schema": e.schema.name, "caption": e.caption}
    )


def make_checksum_entity(
    e: EntityProxy, quiet: bool | None = False
) -> EntityProxy | None:
    """
    Reduce an entity to its ID, schema and contentHash property
    """
    q = bool(quiet)
    stub = make_stub_entity(e)
    if stub is not None:
        stub.add("contentHash", e.get("contentHash", quiet=q), quiet=q)
        return stub



"""
Access legacy Aleph servicelayer archive read-only without `servicelayer`
dendency
"""

from enum import StrEnum

from anystore.functools import weakref_cache as cache
from anystore.store import BaseStore, get_store
from anystore.types import Uri
from pydantic_settings import BaseSettings, SettingsConfigDict

from openaleph_procrastinate.exceptions import ArchiveFileNotFound


def make_checksum_key(ch: str) -> str:
    if len(ch) < 6:
        raise ValueError(f"Invalid checksum: `{ch}`")
    return "/".join((ch[:2], ch[2:4], ch[4:6], ch))


class ArchiveType(StrEnum):
    file = "file"
    s3 = "s3"
    gcs = "gcs"


class ArchiveSettings(BaseSettings):
    model_config = SettingsConfigDict(
        env_prefix="archive_", env_file=".env", extra="ignore"
    )

    type: ArchiveType = ArchiveType.file
    bucket: str | None = None
    path: str | None = None
    endpoint_url: str | None = None


@cache
def get_archive(uri: Uri | None = None) -> BaseStore:
    if uri is not None:
        return get_store(uri=uri)
    settings = ArchiveSettings()
    if settings.type == ArchiveType.s3:
        return get_store(f"s3://{settings.bucket}")
    if settings.type == ArchiveType.gcs:
        return get_store(f"gcs://{settings.bucket}")
    return get_store(uri=settings.path)


def lookup_key(checksum: str, archive: BaseStore | None = None) -> str:
    archive = archive or get_archive()
    prefix = make_checksum_key(checksum)
    for key in archive.iterate_keys(prefix=prefix):
        return key
    raise ArchiveFileNotFound(f"Key does not exist: `{prefix}`")
